import collections
from bson import ObjectId

from .cursor import MontyCursor
from .base import BaseObject
from .engine.queries import QueryFilter
from .engine.update import Updator
from .results import (BulkWriteResult,
                      DeleteResult,
                      InsertOneResult,
                      InsertManyResult,
                      UpdateResult)


def _internal_scan_query(query_spec, collection):
    queryfilter = QueryFilter(query_spec)
    storage = collection.database.client._storage
    documents = storage.query(collection.database.name,
                              collection.name,
                              collection.write_concern,
                              collection.codec_options,
                              0)
    for doc in documents:
        if queryfilter(doc):
            yield queryfilter.field_walker


class MontyCollection(BaseObject):

    def __init__(self, database, name, create=False,
                 codec_options=None, write_concern=None, **kwargs):
        """
        """
        super(MontyCollection, self).__init__(
            codec_options or database.codec_options,
            write_concern or database.write_concern)

        self._database = database
        self._name = name

    def __repr__(self):
        return "MontyCollection({!r}, {!r})".format(self._database, self._name)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (self._database == other.database and
                    self._name == other.name)
        return NotImplemented

    def __ne__(self, other):
        return not self == other

    def __getattr__(self, name):
        if name.startswith('_'):
            full_name = ".".join((self._name, name))
            raise AttributeError(
                "MontyCollection has no attribute {0!r}. To access the {1}"
                " collection, use database[{1!r}].".format(name, full_name))
        return self.__getitem__(name)

    def __getitem__(self, key):
        return self._database.get_collection(".".join((self._name, key)))

    @property
    def full_name(self):
        """
        """
        return u".".join((self._database.name, self._name))

    @property
    def name(self):
        """
        """
        return self._name

    @property
    def database(self):
        """
        """
        return self._database

    def with_options(self, codec_options=None, write_concern=None):
        raise NotImplementedError("Not implemented.")

    def bulk_write(self, requests, ordered=True):
        raise NotImplementedError("Not implemented.")

    def insert_one(self,
                   document,
                   bypass_document_validation=False, *args, **kwargs):
        """
        """
        if bypass_document_validation:
            pass

        if "_id" not in document:
            document["_id"] = ObjectId()

        return InsertOneResult(
            self.database.client._storage.write_one(
                self.database.name,
                self._name,
                self.write_concern,
                self.codec_options,
                document
            ))

    def insert_many(self,
                    documents,
                    ordered=True,
                    bypass_document_validation=False, *args, **kwargs):
        """
        """
        if not isinstance(documents, collections.Iterable) or not documents:
            raise TypeError("documents must be a non-empty list")

        if bypass_document_validation:
            pass

        for doc in documents:
            if "_id" not in doc:
                doc["_id"] = ObjectId()

        return InsertManyResult(
            self.database.client._storage.write_many(
                self.database.name,
                self._name,
                self.write_concern,
                self.codec_options,
                documents,
                ordered
            ))

    def replace_one(self,
                    filter,
                    replacement,
                    upsert=False,
                    bypass_document_validation=False, *args, **kwargs):
        """
        """
        if bypass_document_validation:
            pass

        raise NotImplementedError("Not implemented.")

    def update_one(self,
                   filter,
                   update,
                   upsert=False,
                   bypass_document_validation=False,
                   array_filters=None, *args, **kwargs):
        """
        """
        if bypass_document_validation:
            pass

        updator = Updator(update)
        fw = next(_internal_scan_query(filter, self))
        if updator(fw):
            self.database.client._storage.write_one(
                self.database.name,
                self._name,
                self.write_concern,
                self.codec_options,
                updator.field_walker.doc
            )

        return UpdateResult({
            "n": None,
            "nModified": None,
            "upserted": None,
        })

    def update_many(self,
                    filter,
                    update,
                    upsert=False,
                    bypass_document_validation=False,
                    array_filters=None, *args, **kwargs):
        """
        """

        if bypass_document_validation:
            pass

        updator = Updator(update)
        for fw in _internal_scan_query(filter, self):
            if updator(fw):
                self.database.client._storage.write_one(
                    self.database.name,
                    self._name,
                    self.write_concern,
                    self.codec_options,
                    updator.field_walker.doc
                )

        return UpdateResult({
            "n": None,
            "nModified": None,
            "upserted": None,
        })

    def delete_one(self, filter):
        raise NotImplementedError("Not implemented.")

    def delete_many(self, filter):
        raise NotImplementedError("Not implemented.")

    def aggregate(self, pipeline, session=None, **kwargs):
        # return CommandCursor
        raise NotImplementedError("Not implemented.")

    def aggregate_raw_batches(self, pipeline, **kwargs):
        # return RawBatchCursor
        raise NotImplementedError("Not implemented.")

    def watch(self,
              pipeline=None,
              full_document='default',
              resume_after=None,
              max_await_time_ms=None,
              batch_size=None,
              collation=None,
              session=None):
        # return a changeStreamCursor
        raise NotImplementedError("Not implemented.")

    def find(self, *args, **kwargs):
        # return a cursor
        return MontyCursor(self, *args, **kwargs)

    def find_raw_batches(self, *args, **kwargs):
        # return RawBatchCursor
        raise NotImplementedError("Not implemented.")

    def find_one(self, filter=None, *args, **kwargs):
        """
        """
        if (filter is not None and not
                isinstance(filter, collections.Mapping)):
            filter = {"_id": filter}

        cursor = self.find(filter, *args, **kwargs)
        for result in cursor.limit(-1):
            return result
        return None

    def find_one_and_delete(self, filter, projection=None, sort=None):
        raise NotImplementedError("Not implemented.")

    def find_one_and_replace(self,
                             filter,
                             replacement,
                             projection=None,
                             sort=None, *args, **kwargs):
        raise NotImplementedError("Not implemented.")

    def find_one_and_update(self,
                            filter,
                            update,
                            projection=None,
                            sort=None, *args, **kwargs):
        raise NotImplementedError("Not implemented.")

    def count(self, filter=None):
        raise NotImplementedError("Not implemented.")

    def distinct(self, key, filter=None):
        raise NotImplementedError("Not implemented.")

    def create_index(self, keys):
        raise NotImplementedError("Not implemented.")

    def create_indexes(self, indexes):
        raise NotImplementedError("Not implemented.")

    def drop_index(self, index_or_name):
        raise NotImplementedError("Not implemented.")

    def drop_indexes(self):
        raise NotImplementedError("Not implemented.")

    def reindex(self):
        raise NotImplementedError("Not implemented.")

    def list_indexes(self):
        # return a commandCursor
        raise NotImplementedError("Not implemented.")

    def index_information(self):
        raise NotImplementedError("Not implemented.")

    def drop(self):
        raise NotImplementedError("Not implemented.")

    def rename(self, new_name, session=None, **kwargs):
        raise NotImplementedError("Not implemented.")

    def options(self, session=None):
        raise NotImplementedError("Not implemented.")

    def map_reduce(self,
                   map,
                   reduce,
                   out,
                   full_response=False,
                   session=None, **kwargs):
        raise NotImplementedError("Not implemented.")

    def inline_map_reduce(self,
                          map,
                          reduce,
                          full_response=False,
                          session=None, **kwargs):
        raise NotImplementedError("Not implemented.")

    def parallel_scan(self, num_cursors, session=None, **kwargs):
        raise NotImplementedError("Not implemented.")




from ..errors import WriteError


class Updator(object):

    def __init__(self, spec):

        self.update_ops = {

            # field update ops
            "$inc": parse_inc,
            "$min": None,
            "$max": None,
            "$mul": None,
            "$rename": None,
            "$set": None,
            "$setOnInsert": None,
            "$unset": None,
            "$currentDate": None,

        }

        self.conditions = self.parser(spec)
        self.__field_walker = None

    def __repr__(self):
        pass

    def __call__(self, fieldwalker):
        self.__field_walker = fieldwalker
        return fieldwalker.doc

    @property
    def field_walker(self):
        return self.__field_walker

    def parser(self, spec):
        if not next(iter(spec)).startswith("$"):
            raise ValueError("update only works with $ operators")

        field_to_update = []
        for op, cmd in spec.items():
            if op not in self.update_ops:
                raise WriteError("Unknown modifier: {}".format(op))
            for field in cmd:
                if field in field_to_update:
                    raise WriteError("Cannot update {0!r} and {0!r} at the "
                                     "same time".format(field))
                field_to_update.append(field)


def parse_inc(query):
    def _inc(field_walker):
        return field_walker

    return _inc




from bson.py3compat import string_type

from ..errors import OperationFailure
from .queries import QueryFilter
from .helpers import (
    is_mapping_type,
    is_array_type,
)


def _is_include(val):
    """
    [] and "" will be `True` as well
    """
    return bool(is_array_type(val) or isinstance(val, string_type) or val)


def _is_positional_match(conditions, match_field):
    """
    @conditions `.queries.LogicBox`
    """
    theme = conditions.theme
    if theme.startswith("$"):
        for con in conditions:
            if _is_positional_match(con, match_field):
                return True
        return False
    else:
        if not theme:
            return False
        return match_field.split(".", 1)[0] == theme.split(".", 1)[0]


def _perr_doc(val):
    """
    For pretty error msg, same as Mongo
    """
    v_lis = []
    for _k, _v in val.items():
        if isinstance(_v, string_type):
            v_lis.append("{0}: \"{1}\"".format(_k, _v))
        else:
            if is_mapping_type(_v):
                _v = _perr_doc(_v)
            if is_array_type(_v):
                _ = []
                for v in _v:
                    _.append(_perr_doc(v))
                _v = "[ " + ", ".join(_) + " ]"
            v_lis.append("{0}: {1}".format(_k, _v))
    return "{ " + ", ".join(v_lis) + " }"


class Projector(object):
    """
    """

    ARRAY_OP_NORMAL = 0
    ARRAY_OP_POSITIONAL = 1
    ARRAY_OP_ELEM_MATCH = 2

    def __init__(self, spec, qfilter):
        self.proj_with_id = True
        self.include_flag = None
        self.regular_field = []
        self.array_field = {}

        self.parser(spec, qfilter)

    def __call__(self, field_walker):
        """
        """
        if not self.proj_with_id:
            del field_walker.doc["_id"]

        for field_path in self.array_field:
            self.array_field[field_path](field_walker)

        if self.include_flag:
            self.regular_field += list(self.array_field.keys())
            self.inclusion(field_walker, self.regular_field)
        else:
            self.exclusion(field_walker, self.regular_field)

    def parser(self, spec, qfilter):
        """
        """
        self.array_op_type = self.ARRAY_OP_NORMAL

        for key, val in spec.items():
            # Parsing options
            if is_mapping_type(val):
                if not len(val) == 1:
                    _v = _perr_doc(val)
                    raise OperationFailure(">1 field in obj: {}".format(_v))

                # Array field options
                sub_k, sub_v = next(iter(val.items()))
                if sub_k == "$slice":
                    if isinstance(sub_v, int):
                        if sub_v >= 0:
                            slicing = slice(sub_v)
                        else:
                            slicing = slice(sub_v, None)
                    elif is_array_type(sub_v):
                        if not len(sub_v) == 2:
                            raise OperationFailure("$slice array wrong size")
                        if sub_v[1] <= 0:
                            raise OperationFailure(
                                "$slice limit must be positive")
                        slicing = slice(sub_v[0], sub_v[0] + sub_v[1])
                    else:
                        raise OperationFailure(
                            "$slice only supports numbers and [skip, limit] "
                            "arrays")

                    self.array_field[key] = self.parse_slice(key, slicing)

                elif sub_k == "$elemMatch":
                    if not is_mapping_type(sub_v):
                        raise OperationFailure("elemMatch: Invalid argument, "
                                               "object required.")
                    if self.array_op_type == self.ARRAY_OP_POSITIONAL:
                        raise OperationFailure("Cannot specify positional "
                                               "operator and $elemMatch.")
                    if "." in key:
                        raise OperationFailure(
                            "Cannot use $elemMatch projection on a nested "
                            "field.")

                    self.array_op_type = self.ARRAY_OP_ELEM_MATCH

                    self.array_field[key] = self.parse_elemMatch(key, sub_v)

                elif sub_k == "$meta":
                    # Currently Not supported.
                    raise NotImplementedError("Monty currently not support "
                                              "$meta in projection.")

                else:
                    _v = _perr_doc(val)
                    raise OperationFailure(
                        "Unsupported projection option: "
                        "{0}: {1}".format(key, _v))

            elif key == "_id" and not _is_include(val):
                self.proj_with_id = False

            else:
                # Normal field options, include or exclude.
                flag = _is_include(val)
                if self.include_flag is None:
                    self.include_flag = flag
                else:
                    if not self.include_flag == flag:
                        raise OperationFailure(
                            "Projection cannot have a mix of inclusion and "
                            "exclusion.")

                self.regular_field.append(key)

            # Is positional ?
            bad_ops = [".$ref", ".$id", ".$db"]
            if ".$" in key and not any(ops in key for ops in bad_ops):
                # Validate the positional op.
                if not _is_include(val):
                    raise OperationFailure(
                        "Cannot exclude array elements with the positional "
                        "operator.")
                if self.array_op_type == self.ARRAY_OP_POSITIONAL:
                    raise OperationFailure(
                        "Cannot specify more than one positional proj. "
                        "per query.")
                if self.array_op_type == self.ARRAY_OP_ELEM_MATCH:
                    raise OperationFailure(
                        "Cannot specify positional operator and $elemMatch.")
                if ".$" in key.split(".$", 1)[-1]:
                    raise OperationFailure(
                        "Positional projection '{}' contains the positional "
                        "operator more than once.".format(key))

                path = key.split(".$", 1)[0]
                conditions = qfilter.conditions
                if not _is_positional_match(conditions, path):
                    raise OperationFailure(
                        "Positional projection '{}' does not match the query "
                        "document.".format(key))

                self.array_op_type = self.ARRAY_OP_POSITIONAL

                self.array_field[path] = self.parse_positional(path)

        if self.include_flag is None:
            self.include_flag = False

    def parse_slice(self, field_path, slicing):
        def _slice(field_walker):
            if "$" in field_path:
                return

            if "." in field_path:
                fore_path, key = field_path.rsplit(".", 1)
                if field_walker(fore_path).exists:
                    for emb_doc in field_walker.value:
                        if key not in emb_doc:
                            continue
                        if is_array_type(emb_doc[key]):
                            emb_doc[key] = emb_doc[key][slicing]
            else:
                doc = field_walker.doc
                if field_path in doc:
                    if is_array_type(doc[field_path]):
                        doc[field_path] = doc[field_path][slicing]

        return _slice

    def parse_elemMatch(self, field_path, sub_v):
        qfilter_ = QueryFilter(sub_v)

        def _elemMatch(field_walker):
            doc = field_walker.doc
            has_match = False
            if field_path in doc and is_array_type(doc[field_path]):
                for emb_doc in doc[field_path]:
                    if qfilter_(emb_doc):
                        doc[field_path] = [emb_doc]
                        has_match = True
                        break
            if not has_match:
                del field_walker.doc[field_path]

            if not self.include_flag:
                self.inclusion(field_walker, [field_path])

        return _elemMatch

    def parse_positional(self, field_path):
        def _positional(field_walker):
            matched_index = field_walker.matched_index(field_path)
            value = {}
            field_walker = field_walker(field_path)
            elements = field_walker.value.elements
            match = True
            paths = field_path.split(".")
            paths.reverse()
            doc = field_walker.doc
            path = ""
            while paths:
                path = paths.pop()
                if is_array_type(doc[path]):
                    if matched_index is None:
                        match = False
                    elif matched_index >= len(doc[path]):
                        match = False
                    else:
                        if len(elements) == 0:
                            raise OperationFailure(
                                "Executor error during find command: "
                                "BadValue: positional operator ({}.$) "
                                "requires corresponding field in query "
                                "specifier".format(field_path))
                        value = [doc[path][matched_index]]
                    break
                if not len(paths):
                    value = {}
                    break
                doc = doc[path]
            if match:
                doc[path] = value
            else:
                del doc[path]

        return _positional

    def drop_doc(self, field_walker, key):
        if field_walker.exists:
            for emb_doc in field_walker.value:
                if key in emb_doc:
                    del emb_doc[key]

    def inclusion(self, field_walker, include_field, fore_path=""):
        if fore_path:
            key_list = []
            for val in field_walker.value:
                if is_mapping_type(val):
                    key_list += list(val.keys())
            field_walker.reset()
            key_list = list(set(key_list))
        else:
            key_list = list(field_walker.doc.keys())

        if "_id" in key_list:
            key_list.remove("_id")

        for key in key_list:
            current_path = fore_path + key

            if current_path in include_field:
                # skip included field
                continue

            drop = True
            for field_path in include_field:
                if field_path.startswith(current_path):
                    drop = False
                    break

            if drop:
                if fore_path:
                    with field_walker(fore_path[:-1]):
                        self.drop_doc(field_walker, key)
                else:
                    if key in field_walker.doc:
                        del field_walker.doc[key]
            else:
                fore_path = current_path + "."
                with field_walker(current_path):
                    self.inclusion(field_walker, include_field, fore_path)

    def exclusion(self, field_walker, exclude_field):
        for field_path in exclude_field:
            if "." in field_path:
                fore_path, key = field_path.rsplit(".", 1)
                with field_walker(fore_path):
                    self.drop_doc(field_walker, key)
            else:
                if field_path in field_walker.doc:
                    del field_walker.doc[field_path]




import os
import shutil
import sqlite3
import contextlib

from bson import BSON
from bson.py3compat import _unicode

from ..base import WriteConcern
from .base import (
    StorageConfig,
    AbstractStorage,
    AbstractDatabase,
    AbstractCollection,
    AbstractCursor,
)


"""
(NOTE) SQLite3 pragmas DEFAULT value:

* Applies to the database
    - journal_mode=delete

* Applies to the connection
    - busy_timeout=5000
    - foreign_keys=OFF
    - automatic_index=ON
"""


SQLITE_CONFIG = """
storage:
  engine: SQLiteStorage
  config: SQLiteConfig
  module: {}
connection:
  journal_mode: WAL
write_concern:
  # These will be picked up by wconcern
  synchronous: 1
  automatic_index: OFF
  busy_timeout: 5000
""".format(__name__)


SQLITE_CONFIG_SCHEMA = """
type: object
required:
  - connection
  - write_concern
properties:
  connection:
    type: object
    properties:
      journal_mode:
        type: string
        enum: [DELETE, TRUNCATE, PERSIST, MEMORY, WAL, "OFF"]
  write_concern:
    type: object
    properties:
      synchronous:
        oneOf:
          - type: string
            enum: ["OFF", NORMAL, FULL, EXTRA, "0", "1", "2", "3"]
          - type: integer
            enum: [0, 1, 2, 3]
      automatic_index:
        oneOf:
          - type: boolean
          - type: string
            enum: ["ON", "OFF"]
      busy_timeout:
        type: integer
"""


class SQLiteConfig(StorageConfig):
    """SQLite storage configuration settings

    Default configuration and schema of SQLite storage
    """
    config = SQLITE_CONFIG
    schema = SQLITE_CONFIG_SCHEMA


SQLITE_DB_EXT = ".collection"
SQLITE_RECORD_TABLE = "documents"


"""SQL"""

CREATE_TABLE = """
    CREATE TABLE [{}](
        k text NOT NULL,
        v blob NOT NULL,
        PRIMARY KEY(k)
    );
"""

INSERT_RECORD = """
    INSERT INTO [{}](k, v) VALUES (?, ?)
"""

SELECT_ALL_RECORD = """
    SELECT v FROM [{}]
"""

SELECT_LIMIT_RECORD = """
    SELECT v FROM [{0}] LIMIT {1}
"""


class SQLiteKVEngine(object):

    def __init__(self, db_pragmas):
        self.__db_pragmas = db_pragmas
        self.__conn = None

    @property
    def db_pragmas(self):
        return self._assemble_pragmas(self.__db_pragmas)

    def _connect(self, db_file, wconcern=None):
        self.__conn = sqlite3.connect(db_file)
        self.__conn.text_factory = str

        wcon_pragmas = ""
        if wconcern:
            wcon_doc = wconcern.document
            # wtimeout (milliseconds) -> busy_timeout (milliseconds)
            timeout = wcon_doc.get("wtimeout")
            if timeout:
                del wcon_doc["wtimeout"]
                wcon_doc["busy_timeout"] = timeout

            wcon_pragmas = self._assemble_pragmas(wcon_doc)

        # update connection pragmas and write_concern pragmas
        self.__conn.executescript(self.db_pragmas + ";" + wcon_pragmas)

        return contextlib.closing(self.__conn)

    def _assemble_pragmas(self, pragma_dict):
        return ";".join(["PRAGMA {0}={1}".format(k, v)
                         for k, v in pragma_dict.items()])

    def create_table(self, db_file):
        with self._connect(db_file) as conn:
            with conn:
                conn.execute(CREATE_TABLE.format(SQLITE_RECORD_TABLE))

    def write_one(self, db_file, params, wconcern=None):
        with self._connect(db_file, wconcern) as conn:
            with conn:
                sql = INSERT_RECORD.format(SQLITE_RECORD_TABLE)
                conn.execute(sql, params)

    def write_many(self, db_file, seq_params, wconcern=None):
        with self._connect(db_file, wconcern) as conn:
            with conn:
                sql = INSERT_RECORD.format(SQLITE_RECORD_TABLE)
                conn.executemany(sql, seq_params)

    def read_all(self, db_file, limit):
        if not os.path.isfile(db_file):
            return []
        with self._connect(db_file) as conn:
            with conn:
                if limit:
                    sql = SELECT_LIMIT_RECORD.format(
                        SQLITE_RECORD_TABLE, limit)
                else:
                    sql = SELECT_ALL_RECORD.format(SQLITE_RECORD_TABLE)

                return conn.execute(sql).fetchall()


class SQLiteWriteConcern(WriteConcern):
    """
    """

    def __init__(self,
                 wtimeout=None,
                 synchronous=None,
                 automatic_index=None):

        super(SQLiteWriteConcern, self).__init__(wtimeout)

        if synchronous is not None:
            self._document["synchronous"] = synchronous
        if automatic_index is not None:
            self._document["automatic_index"] = automatic_index

    def __repr__(self):
        return ("SQLiteWriteConcern({})".format(
            ", ".join("%s=%s" % kvt for kvt in self.document.items()),))


class SQLiteStorage(AbstractStorage):
    """
    """

    def __init__(self, repository, storage_config):
        super(SQLiteStorage, self).__init__(repository, storage_config)
        self._conn = SQLiteKVEngine(self._config.connection)

    def _db_path(self, db_name):
        """
        Get Monty database dir path.
        """
        return os.path.join(self._repository, db_name)

    def wconcern_parser(self, client_kwargs):
        _client_btimeout = client_kwargs.get("busy_timeout")
        # Default from config
        wcon_pragmas = self._config.write_concern
        wtimeout = client_kwargs.get(
            "wtimeout",
            _client_btimeout or wcon_pragmas.get("busy_timeout"))
        synchronous = client_kwargs.get(
            "synchronous",
            wcon_pragmas.get("synchronous"))
        automatic_index = client_kwargs.get(
            "automatic_index",
            wcon_pragmas.get("automatic_index"))

        return SQLiteWriteConcern(wtimeout,
                                  synchronous,
                                  automatic_index)

    def database_create(self, db_name):
        if not os.path.isdir(self._db_path(db_name)):
            os.makedirs(self._db_path(db_name))

    def database_drop(self, db_name):
        db_path = self._db_path(db_name)
        if os.path.isdir(db_path):
            shutil.rmtree(db_path)

    def database_list(self):
        return [
            name for name in os.listdir(_unicode(self._repository))
            if os.path.isdir(self._db_path(name))
        ]


class SQLiteDatabase(AbstractDatabase):

    def __init__(self, name, storage):
        super(SQLiteDatabase, self).__init__(name, storage)
        self._db_path = storage._db_path(self._name)

    def _col_path(self, col_name):
        """
        Get SQLite database file path, which is Monty collection.
        """
        return os.path.join(self._db_path, col_name) + SQLITE_DB_EXT

    @property
    def _conn(self):
        return self._storage._conn

    def database_exists(self):
        return os.path.isdir(self._db_path)

    def collection_exists(self, col_name):
        return os.path.isfile(self._col_path(col_name))

    def collection_create(self, col_name):
        if not self.database_exists():
            self._storage.database_create(self._name)
        self._conn.create_table(self._col_path(col_name))

    def collection_drop(self, col_name):
        if self.collection_exists(col_name):
            os.remove(self._col_path(col_name))

    def collection_list(self):
        if not self.database_exists():
            return []
        return [os.path.splitext(name)[0]
                for name in os.listdir(_unicode(self._db_path))]


SQLiteStorage.db_cls = SQLiteDatabase


class SQLiteCollection(AbstractCollection):

    def __init__(self, name, database, write_concern, codec_options):
        super(SQLiteCollection, self).__init__(name,
                                               database,
                                               write_concern,
                                               codec_options)

        self._col_path = self._database._col_path(self._name)

    def _ensure_table(func):
        def make_table(self, *args, **kwargs):
            if not self._database.collection_exists(self._name):
                self._database.collection_create(self._name)
            return func(self, *args, **kwargs)
        return make_table

    def _encode_doc(self, doc):
        # Preserve BSON types
        encoded = BSON.encode(doc, False, self.coptions)
        return sqlite3.Binary(encoded)

    @property
    def _conn(self):
        return self._database._conn

    @_ensure_table
    def write_one(self, doc):
        """
        """
        self._conn.write_one(
            self._col_path,
            (str(doc["_id"]), self._encode_doc(doc),),
            self.wconcern
        )

        return doc["_id"]

    @_ensure_table
    def write_many(self, docs, ordered=True):
        """
        """
        self._conn.write_many(
            self._col_path,
            [(str(doc["_id"]), self._encode_doc(doc)) for doc in docs],
            self.wconcern
        )

        return [doc["_id"] for doc in docs]


SQLiteDatabase.col_cls = SQLiteCollection


class SQLiteCursor(AbstractCursor):

    def __init__(self, collection):
        super(SQLiteCursor, self).__init__(collection)

    @property
    def _conn(self):
        return self._collection._conn

    @property
    def _col_path(self):
        return self._collection._col_path

    def _decode_doc(self, doc):
        # Decode BSON types
        return BSON(doc[0]).decode(self._collection.coptions)

    def query(self, max_scan):
        docs = self._conn.read_all(self._col_path, max_scan)
        return [self._decode_doc(doc) for doc in docs]


SQLiteCollection.cursor_cls = SQLiteCursor




from bson import SON, BSON

from .base import (
    StorageConfig,
    AbstractStorage,
    AbstractDatabase,
    AbstractCollection,
    AbstractCursor,
)


MEMORY_CONFIG = """
storage:
  engine: MemoryStorage
  config: MemoryConfig
  module: {}
""".format(__name__)


class MemoryConfig(StorageConfig):
    """Memory storage configuration settings

    Default configuration of Memory storage
    """
    config = MEMORY_CONFIG
    schema = None


MEMORY_REPOSITORY = ":memory:"


class MemoryStorage(AbstractStorage):
    """
    """

    def __init__(self, repository, storage_config):
        super(MemoryStorage, self).__init__(repository, storage_config)
        self._repo = SON()

    def database_create(self, db_name):
        self._repo[db_name] = SON()

    def database_drop(self, db_name):
        if db_name in self._repo:
            del self._repo[db_name]

    def database_list(self):
        return list(self._repo.keys())


class MemoryDatabase(AbstractDatabase):
    """
    """

    @property
    def _db(self):
        return self._storage._repo[self._name]

    def db_exists(self):
        return self._name in self._storage._repo

    def collection_exists(self, col_name):
        if self.db_exists():
            return col_name in self._db
        return False

    def collection_create(self, col_name):
        if not self.db_exists():
            self._storage.database_create(self._name)
        self._db[col_name] = SON()

    def collection_drop(self, col_name):
        if self.collection_exists(col_name):
            del self._db[col_name]

    def collection_list(self):
        if not self.db_exists():
            return []
        return list(self._db.keys())


MemoryStorage.db_cls = MemoryDatabase


class MemoryCollection(AbstractCollection):
    """
    """

    @property
    def _col(self):
        if not self._col_exists():
            self._database.collection_create(self._name)
        return self._database._db[self._name]

    def _col_exists(self):
        return self._database.collection_exists(self._name)

    def _encode_doc(self, doc):
        return BSON.encode(doc, False, self.coptions)

    def write_one(self, doc):
        self._col[str(doc["_id"])] = self._encode_doc(doc)
        return doc["_id"]

    def write_many(self, docs, ordered=True):
        for doc in docs:
            self._col[str(doc["_id"])] = self._encode_doc(doc)
        return [doc["_id"] for doc in docs]


MemoryDatabase.col_cls = MemoryCollection


class MemoryCursor(AbstractCursor):
    """
    """

    @property
    def _col(self):
        if self._collection._col_exists():
            return self._collection._col
        return SON()

    def _decode_doc(self, doc):
        return BSON(doc).decode(self._collection.coptions)

    def query(self, max_scan):
        docs = [self._decode_doc(doc) for doc in self._col.values()]
        if not max_scan:
            return docs
        else:
            return docs[:max_scan]


MemoryCollection.cursor_cls = MemoryCursor




from collections import deque, OrderedDict, Mapping


def is_array_type_(doc):
    return isinstance(doc, (list, tuple, _FieldValues))


def is_mapping_type_(obj):
    return isinstance(obj, Mapping)


class FieldWalker(object):
    """Document traversal context manager

    Attributes:
        doc (Mapping type): The document being query.

        value (_FieldValues): An instance of `_FieldValues`, hold the result
                              of the query.

        exists (bool): Field path exists.

        embedded_in_array (bool): Documents field embedded in array

        index_posed (bool): Path of this query ends with index position.

        array_field_missing (bool): Documents in array has missing field.
            Possible some of document embedded in array has field missing,
            or all document in array has no such field existed.

        array_status_normal (bool): No missing field but the path not exists
            Query path not existed in document array due to array out of index
            or has no document object in array.
            In this case, the field was not missing because it's possible that
            field did not meant to be there (because no embedded document) or
            array out of index if the path is index position, and since the
            field was not missing, the document won't pop when querying `None`.
    """

    __slots__ = [
        "doc",
        "value",
        "exists",
        "embedded_in_array",
        "index_posed",
        "array_field_missing",
        "array_status_normal",

        # FLAGS_FOR_NONE_QUERYING
        "_NQF_been_in_array",
        "_NQF_docs_field_missing_in_array",
        "_NQF_array_field_not_exists_in_all_elements",
        "_NQF_out_of_array_index",
        "_NQF_no_docs_in_array",

        "_matched_indexes",
        "_elem_iter_map",
        "_query_path",
        "_end_key",
    ]

    def __init__(self, doc):
        """
        Args:
            doc (dict): Document object
        """
        self.doc = doc
        self._matched_indexes = dict()
        self.reset()

    def reset(self, partial=None):
        """Rest all, or keeping some flags for internal use.

        Args:
            partial (bool): Reset partial attritubes.
        """
        self.value = _FieldValues()
        self.exists = False
        self.embedded_in_array = False
        self.index_posed = False
        self._elem_iter_map = OrderedDict()
        self._query_path = ""
        self._end_key = None

        if not partial:
            self._NQF_been_in_array = False
            self._NQF_docs_field_missing_in_array = False
            self._NQF_array_field_not_exists_in_all_elements = False
            self._NQF_out_of_array_index = False
            self._NQF_no_docs_in_array = False

    def __call__(self, path):
        """Walk through document and acquire value with given key-path

        Args:
            path (string): Document field path
        """
        doc_ = self.doc
        ref_ = None
        end_key_ = None
        self.reset()

        self._query_path = path
        for field in path.split("."):

            field_as_index = False
            array_has_doc = False
            if is_array_type_(doc_):
                if len(doc_) == 0:
                    self.exists = False
                    break

                self._NQF_been_in_array = True
                array_has_doc = any(is_mapping_type_(e_) for e_ in doc_)
                field_as_index = field.isdigit()

                if field_as_index:
                    if self.index_posed and self.embedded_in_array:
                        field_as_index = any(is_array_type_(e_) for e_ in doc_)
                else:
                    doc_ = self._walk_array(doc_, field)

            # Is the path ends with index position ?
            self.index_posed = field_as_index

            # If the `doc_` is an array (or `_FieldValues` type) and containing
            # documents, those documents possible has numeric string key,
            # for example: [{"1": <value>}, ...]
            if field_as_index and array_has_doc:
                # Index position path As a Field of `doc_`
                iaf_doc_ = self._walk_array(doc_, field)
                # Append index position result to the document field result
                if iaf_doc_ is not None:
                    if len(doc_) > int(field):  # Make sure index in range
                        if isinstance(doc_, _FieldValues):
                            iaf_doc_[field] += doc_._positional(int(field))
                        else:
                            iaf_doc_[field]._append(doc_[int(field)])

                    doc_ = iaf_doc_
                    field_as_index = False

            if field_as_index and self.embedded_in_array:
                # the `doc_` in here must be `_FieldValues` type
                doc_ = {field: doc_._positional(int(field))}
                field_as_index = False

            if doc_ is not None and self.embedded_in_array:
                # `doc_` should be {field: _FieldValues},
                # not a bare _FieldValues instance.
                if isinstance(doc_, _FieldValues):
                    ref_ = None
                else:
                    ref_ = doc_[field].ref
            else:
                ref_ = doc_
            end_key_ = field

            try:
                # Try getting value with key(field) or index.
                doc_ = doc_[int(field) if field_as_index else field]
                self.exists = True

            except (KeyError, IndexError, TypeError) as err:
                ecls = err.__class__
                # Raising some flags if conditions match.

                # FLAGS_FOR_NONE_QUERYING:
                #   possible index position out of length of array
                self._NQF_out_of_array_index = ecls is IndexError
                # FLAGS_FOR_NONE_QUERYING:
                #   possible not field missing, but the array has no document
                if ecls is TypeError and self._NQF_been_in_array:
                    self._NQF_no_docs_in_array = (
                        not self._NQF_docs_field_missing_in_array)

                # Reset partialy and stop field walking.
                doc_ = None
                ref_ = None
                end_key_ = None
                self.reset(partial=True)
                break

        """End of walk"""

        # Collecting values
        if not field_as_index and is_array_type_(doc_):
            # Extend `fieldValues.elements` with an array field value from
            # a single document or from multiple documents inside an array.
            self.value._extend(doc_)
        # Append to `fieldValues.arrays`, but if `doc_` is not array type,
        # will be append to `fieldValues.elements`.
        self.value._append(doc_)

        self.value.ref = ref_
        self._end_key = end_key_

        # FLAGS_FOR_NONE_QUERYING:
        #   Correcting flag after value been collected.
        #       Confirm all documents inside the array have no such field,
        #       instead of missing field in some of the documents.
        if (None not in self.value.elements and
                not self._NQF_array_field_not_exists_in_all_elements):
            self._NQF_docs_field_missing_in_array = False

        self.array_field_missing = self._NQF_docs_field_missing_in_array
        self.array_status_normal = (self._NQF_out_of_array_index or
                                    self._NQF_no_docs_in_array)
        return self

    def _walk_array(self, doc_, field):
        """Walk in to array for embedded documents.
        """
        field_values = _FieldValues()
        ref_ = []
        num_of_emb_doc = 0
        self._elem_iter_map[field] = OrderedDict()

        for i, emb_doc in enumerate(doc_):
            if not is_mapping_type_(emb_doc):
                continue
            num_of_emb_doc += 1

            emb_field = FieldWalker(emb_doc)(field)
            if emb_field.exists:
                self._elem_iter_map[field][i] = len(emb_field.value.elements)
                field_values += emb_field.value
                ref_.append(emb_field.value.ref)
            else:
                # FLAGS_FOR_NONE_QUERYING:
                #   field not exists in all elements.
                self._NQF_array_field_not_exists_in_all_elements = True

        if len(field_values.arrays) != num_of_emb_doc:
            # FLAGS_FOR_NONE_QUERYING:
            #   Possible missing field in some documents.
            #       Using `field_values.arrays` length to compare is not
            #       accurate, but will correcting the result after all value
            #       been collected.
            self._NQF_docs_field_missing_in_array = True

        if field_values:
            self.embedded_in_array = True
            field_values.ref = ref_
            return {field: field_values}
        else:
            return None

    def _get_matched_index(self):
        times = self.value.iter_times
        if len(self._elem_iter_map) == 0:
            return None if len(self.value.elements) == 0 else (times - 1)
        else:
            while len(self._elem_iter_map):
                for ind, len_ in self._elem_iter_map.popitem()[1].items():
                    if times > len_:
                        times -= len_
                    else:
                        times = ind + 1
                        break
            return times - 1

    def __enter__(self):
        return self

    def __exit__(self, *args):
        root = self._query_path.split(".", 1)[0]
        self._matched_indexes[root] = self._get_matched_index()
        self.reset(partial=True)

    def matched_index(self, path):
        """
        """
        return self._matched_indexes.get(path.split(".", 1)[0])

    def setval(self, value):
        if self.value.ref is None:
            return

        ref_ = self.value.ref
        if not self.embedded_in_array:
            ref_ = [ref_]

        for r_ in ref_:
            if is_array_type_(r_):
                if len(r_) > int(self._end_key):
                    r_[int(self._end_key)] = value
                else:
                    fill = int(self._end_key) - len(r_)
                    r_ += [None for i in range(fill)] + [value]
            elif is_mapping_type_(r_):
                r_[self._end_key] = value


class _FieldValues(object):

    __slots__ = [
        "elements",
        "arrays",
        "iter_queue",
        "iter_times",
        "ref",
    ]

    def __init__(self, elements=None, arrays=None):
        self.elements = elements or []
        self.arrays = arrays or []
        self.iter_queue = deque()
        self.iter_times = 1
        self.ref = None

    def _merged(self):
        return self.elements + self.arrays

    def __repr__(self):
        return "_FieldValues(elements: {}, arrays: {})".format(self.elements,
                                                               self.arrays)

    def __next__(self):
        if len(self.iter_queue):
            self.iter_times += 1
            return self.iter_queue.popleft()
        else:
            raise StopIteration

    next = __next__

    def __iter__(self):
        self.iter_times = 0
        self.iter_queue = deque(self._merged())
        return self

    def __len__(self):
        return len(self._merged())

    def __eq__(self, other):
        return self._merged() == other

    def __bool__(self):
        return bool(self.elements or self.arrays)

    __nonzero__ = __bool__

    def __getitem__(self, index):
        return self.elements[index]

    def __iadd__(self, val):
        self.elements += val.elements
        self.arrays += val.arrays
        return self

    def _extend(self, val):
        if isinstance(val, _FieldValues):
            self.elements += val.elements
        else:
            self.elements += val

    def _append(self, val):
        if isinstance(val, _FieldValues):
            self.arrays += val.arrays
        else:
            if is_array_type_(val):
                self.arrays.append(val)
            else:
                self.elements.append(val)

    def _positional(self, index):
        self.elements = [val[index] for val in self.arrays
                         if len(val) > index]
        self.arrays = []

        self.ref = [list(r.values())[0] for r in self.ref]

        return self




import sys
import re
from collections import Mapping


FS_ENCODE = sys.getfilesystemencoding()

PY3 = sys.version_info[0] == 3
PY36 = sys.version_info[0] == 3 and sys.version_info[1] >= 6

RE_PATTERN_TYPE = type(re.compile(""))


def is_mapping_type(obj):
    return isinstance(obj, Mapping)


def is_array_type(obj):
    return isinstance(obj, (list, tuple))


def is_pattern_type(obj):
    return isinstance(obj, RE_PATTERN_TYPE)


def re_int_flag_to_str(int_flags):
    """Ripped from bson.json_util
    """
    flags = ""
    if int_flags & re.IGNORECASE:
        flags += "i"
    if int_flags & re.LOCALE:
        flags += "l"
    if int_flags & re.MULTILINE:
        flags += "m"
    if int_flags & re.DOTALL:
        flags += "s"
    if int_flags & re.UNICODE:
        flags += "u"
    if int_flags & re.VERBOSE:
        flags += "x"
    return flags


def decode_(s):
    if PY3:
        return s
    else:
        return s.decode(FS_ENCODE)


def encode_(s):
    if PY3:
        return s
    else:
        return s.encode("utf-8")


# Grab from six.py
def with_metaclass(meta, *bases):
    """Create a base class with a metaclass."""
    # This requires a bit of explanation: the basic idea is to make a dummy
    # metaclass for one level of class instantiation that replaces itself with
    # the actual metaclass.
    class metaclass(meta):

        def __new__(cls, name, this_bases, d):
            return meta(name, bases, d)
    return type.__new__(metaclass, 'temporary_class', (), {})


def __add_attrib(deco):
    """Decorator helper"""
    def meta_decorator(value):
        def add_attrib(func):
            func._keep = lambda: value
            return func
        return add_attrib
    return meta_decorator


@__add_attrib
def keep(query):
    """A decorator that preserve operation query for operator"""
    def _(func):
        return query
    return _




import os
import shutil
from bson import BSON, SON
from bson.py3compat import _unicode
from bson.json_util import (
    CANONICAL_JSON_OPTIONS,
    loads as _loads,
    dumps as _dumps,
)

from .base import (
    StorageConfig,
    AbstractStorage,
    AbstractDatabase,
    AbstractCollection,
    AbstractCursor,
)


FALTFILE_CONFIG = """
storage:
  engine: FlatFileStorage
  config: FlatFileConfig
  module: {}
connection:
  cache_modified: 0
""".format(__name__)


FLATFILE_CONFIG_SCHEMA = """
type: object
required:
    - connection
properties:
  connection:
    type: object
    required:
        - cache_modified
    properties:
      cache_modified:
        type: integer
        minimum: 0
"""


class FlatFileConfig(StorageConfig):
    """FaltFile storage configuration settings

    Default configuration and schema of FaltFile storage
    """
    config = FALTFILE_CONFIG
    schema = FLATFILE_CONFIG_SCHEMA


FLATFILE_DB_EXT = ".json"


def _read_pretty(file_path):
    with open(file_path, "r") as fp:
        lines = [line.strip() for line in fp.readlines()]
        if lines:
            serialized = "".join(lines)
            return _loads(serialized, json_options=CANONICAL_JSON_OPTIONS)
        else:
            return []


def _write_pretty(file_path, documents):
    serialized = []
    for doc in documents.values():
        serialized.append(
            _dumps(BSON(doc).decode(), json_options=CANONICAL_JSON_OPTIONS))

    with open(file_path, "w") as fp:
        fp.write("[\n{}\n]".format(",\n".join(serialized)))


class FlatFileKVEngine(object):

    def __init__(self, file_path, conn_config):
        """
        """
        self.file_path = file_path

        self.__conn_config = conn_config
        self.__cache = SON()
        self.modified_count = 0

        if os.path.isfile(self.file_path):
            for doc in _read_pretty(self.file_path):
                self.__cache[doc["_id"]] = BSON.encode(doc)

    @classmethod
    def touch(cls, file_path):
        if not os.path.isfile(file_path):
            with open(file_path, "w"):
                pass

    @property
    def document_count(self):
        return len(self.__cache)

    def flush(self):
        _write_pretty(self.file_path, self.__cache)
        self.modified_count = 0

    def read(self):
        return self.__cache

    def write(self, documents):
        if not isinstance(documents, SON):
            raise TypeError("Expecting 'SON' type, got {!r}"
                            "".format(type(documents).__name__))
        self.modified_count += len(documents)
        self.__cache.update(documents)

        if self.modified_count > self.__conn_config.cache_modified:
            self.flush()

    def delete(self, doc_id):
        self.modified_count += 1
        del self.__cache[doc_id]

        if self.modified_count > self.__conn_config.cache_modified:
            self.flush()


class FlatFileStorage(AbstractStorage):
    """
    """

    def __init__(self, repository, storage_config):
        super(FlatFileStorage, self).__init__(repository, storage_config)
        self._cache_manager = {}

    def _db_path(self, db_name):
        """
        Get Monty database dir path.
        """
        return os.path.join(self._repository, db_name)

    def close(self):
        for db in self._cache_manager:
            for col in self._cache_manager[db]:
                self._cache_manager[db][col].flush()
        self.is_opened = False

    def database_create(self, db_name):
        if not os.path.isdir(self._db_path(db_name)):
            os.makedirs(self._db_path(db_name))

    def database_drop(self, db_name):
        db_path = self._db_path(db_name)
        if os.path.isdir(db_path):
            shutil.rmtree(db_path)

    def database_list(self):
        return [
            name for name in os.listdir(_unicode(self._repository))
            if os.path.isdir(self._db_path(name))
        ]


class FlatFileDatabase(AbstractDatabase):
    """
    """

    def __init__(self, name, storage):
        super(FlatFileDatabase, self).__init__(name, storage)
        if name not in storage._cache_manager:
            storage._cache_manager[name] = {}
        self._db_path = storage._db_path(self._name)

    @property
    def _cache_manager(self):
        return self._storage._cache_manager[self._name]

    def _col_path(self, col_name):
        """
        Get JSON database file path, which is Monty collection.
        """
        return os.path.join(self._db_path, col_name) + FLATFILE_DB_EXT

    def database_exists(self):
        return os.path.isdir(self._db_path)

    def collection_exists(self, col_name):
        return os.path.isfile(self._col_path(col_name))

    def collection_create(self, col_name):
        if not self.database_exists():
            self._storage.database_create(self._name)
        FlatFileKVEngine.touch(self._col_path(col_name))

    def collection_drop(self, col_name):
        if self.collection_exists(col_name):
            os.remove(self._col_path(col_name))

    def collection_list(self):
        if not self.database_exists():
            return []
        return [os.path.splitext(name)[0]
                for name in os.listdir(_unicode(self._db_path))]


FlatFileStorage.db_cls = FlatFileDatabase


class FlatFileCollection(AbstractCollection):
    """
    """

    def __init__(self, name, database, write_concern, codec_options):
        config = database._storage._config
        super(FlatFileCollection, self).__init__(name,
                                                 database,
                                                 write_concern,
                                                 codec_options)

        self._col_path = self._database._col_path(self._name)
        if self._name not in database._cache_manager:
            database._cache_manager[self._name] = FlatFileKVEngine(
                self._col_path, config.connection)

    @property
    def _flatfile(self):
        return self._database._cache_manager[self._name]

    def _ensure_table(func):
        def make_table(self, *args, **kwargs):
            if not self._database.collection_exists(self._name):
                self._database.collection_create(self._name)
            return func(self, *args, **kwargs)
        return make_table

    def _encode_doc(self, doc):
        return BSON.encode(doc, False, self.coptions)

    @_ensure_table
    def write_one(self, doc):
        _doc = SON()
        _doc[doc["_id"]] = self._encode_doc(doc)
        self._flatfile.write(_doc)

        return doc["_id"]

    @_ensure_table
    def write_many(self, docs, ordered=True):
        _docs = SON()
        for doc in docs:
            _docs[doc["_id"]] = self._encode_doc(doc)

        self._flatfile.write(_docs)

        return [doc["_id"] for doc in docs]


FlatFileDatabase.col_cls = FlatFileCollection


class FlatFileCursor(AbstractCursor):
    """
    """

    def __init__(self, collection):
        super(FlatFileCursor, self).__init__(collection)

    @property
    def _flatfile(self):
        return self._collection._flatfile

    def _decode_doc(self, doc):
        # Decode BSON types
        return BSON(doc).decode(self._collection.coptions)

    def query(self, max_scan):
        cache = self._flatfile.read()
        docs = [self._decode_doc(doc) for doc in cache.values()]
        if not max_scan:
            return docs
        else:
            return docs[:max_scan]


FlatFileCollection.cursor_cls = FlatFileCursor



from .base import (
    ASCENDING,
    DESCENDING,
)
from .client import MontyClient
from .database import MontyDatabase
from .collection import MontyCollection
from .cursor import (
    MontyCursor,
    CursorType
)
from .configure import MontyConfigure

from . import utils


__all__ = [

    "MontyClient",
    "MontyDatabase",
    "MontyCollection",
    "MontyCursor",

    "MontyConfigure",

    "ASCENDING",
    "DESCENDING",
    "CursorType",

    "utils",
]



import os
import importlib
import json
import inspect
from jsonschema.validators import Draft4Validator
from collections import MutableMapping, OrderedDict

import yaml
from yaml import SafeLoader, SafeDumper
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper

from .storage.base import StorageConfig
from .storage.memory import MemoryConfig, MEMORY_REPOSITORY
from .storage.sqlite import SQLiteConfig
from .errors import ConfigurationError


BASE_SCHEMA = """
type: object
required:
    - storage
properties:
  storage:
    type: object
    required:
        - engine
        - config
        - module
    properties:
      engine:
        type: string
      config:
        type: string
      module:
        type: string
"""


def _make_validator(schema):
    return Draft4Validator(schema, types={"object": MutableMapping})


def _config_validator(config, config_schema):
    errors = []
    base_val = _make_validator(yaml.load(BASE_SCHEMA, SafeLoader))
    errors += base_val.iter_errors(config)
    if config_schema:
        conf_val = _make_validator(yaml.load(config_schema, SafeLoader))
        errors += conf_val.iter_errors(config)

    header = """
    Configuration file schema error, total {}:
    """.format(len(errors))
    msg_ = """
    - ERR #{}: {}
    config path: {}
    bad value: {}
    schema: {}
    """
    errmsg = []
    for i, e in enumerate(errors, 1):
        p_ = ".".join(e.path)
        errmsg.append(msg_.format(i, e.message, p_, e.instance, e.schema))
    if errmsg:
        raise ConfigurationError("".join([header] + errmsg))


class AttribDict(MutableMapping):
    """
    Based on source:
    Thanks @nivk
    https://stackoverflow.com/a/47081357/4145300
    """

    def __init__(self, ordered):
        super(AttribDict, self).__setattr__('cnf', OrderedDict(ordered))

    def __repr__(self, indent=None):
        return json.dumps(self.cnf, default=OrderedDict, indent=indent)

    def __getattr__(self, key):
        return self.__getitem__(key)

    def __setattr__(self, key, val):
        self.__setitem__(key, val)

    def __delitem__(self, key):
        raise ConfigurationError("Can not delete option.")

    def __getitem__(self, key):
        if key not in self.cnf:
            raise ConfigurationError("Option {!r} does not exists."
                                     "".format(key))
        return self.cnf[key]

    def __setitem__(self, key, val):
        if key not in self.cnf:
            raise ConfigurationError("Adding new option is not allowed.")
        self.cnf[key] = val

    def __iter__(self):
        return iter(self.cnf)

    def __len__(self):
        return len(self.cnf)

    def __restriction__(self, *args, **kwargs):
        raise ConfigurationError('Can not use this method.')

    clear = __restriction__
    pop = __restriction__
    popitem = __restriction__
    setdefault = __restriction__
    update = __restriction__

    def pretty(self):
        return self.__repr__(indent=4)  # pragma: no cover

    def reload(self, repository=None, update=None):
        if repository is not None:
            configure = MontyConfigure(repository)
            if not configure.in_memory:
                if not configure.exists():
                    raise ConfigurationError("Config file must be saved "
                                             "before reload.")
                update = configure.load().config

        if update is not None:
            for key in self.cnf:
                # new config should have the same key.
                val = update[key]
                if isinstance(self.cnf[key], AttribDict):
                    self.cnf[key].reload(update=val)
                else:
                    self.cnf[key] = val


def yaml_config_load(stream, Loader=Loader, object_pairs_hook=AttribDict):
    """
    Based on source:
    Thanks @coldfix
    https://stackoverflow.com/a/21912744/4145300
    """
    class AttribLoader(Loader):
        pass

    def construct_mapping(loader, node):
        loader.flatten_mapping(node)
        return object_pairs_hook(loader.construct_pairs(node))

    AttribLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
        construct_mapping)

    return yaml.load(stream, AttribLoader)


def yaml_config_dump(data, stream=None, Dumper=Dumper, **kwds):
    """
    Based on source:
    Thanks @coldfix.
    https://stackoverflow.com/a/21912744/4145300
    """
    class AttribDumper(Dumper):
        pass

    def _dict_representer(dumper, data):
        return dumper.represent_mapping(
            yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
            data.items())

    AttribDumper.add_representer(AttribDict, _dict_representer)

    return yaml.dump(data, stream, AttribDumper, **kwds)


class MontyConfigure(object):
    """
    Args:
        repository (str): Database root directory path.

        storage_config (str, optional): YAML format config string.
            Defaults to `montydb.storage.SQLiteConfig`.
            If no `conf.yaml` config file exists in `repository` path,
            will take this parameter and save as `conf.yaml`.
            Ignored if `conf.yaml` exists.
    """

    __config_filename = "conf.yaml"

    def __init__(self, repository):
        self.in_memory = repository == MEMORY_REPOSITORY
        self._repository = repository
        self._config = None
        self._schema = None

    def __repr__(self):
        return "MontyConfigure(\n{})".format(self.to_yaml())

    def __enter__(self):
        return self

    def __exit__(self, *args):
        if self._config is not None:
            self.save()

    def _get_storage_engine(self):
        """Get storage engine from config
        """
        engine_cls_name = self._config.storage.engine
        module = importlib.import_module(self._config.storage.module)
        engine_cls = getattr(module, engine_cls_name)
        return engine_cls(self._repository, self._config)

    def _get_storage_config_schema(self):
        """Get storage engine config schema from config
        """
        config_cls_name = self._config.storage.config
        module = importlib.import_module(self._config.storage.module)
        config_cls = getattr(module, config_cls_name)
        return config_cls.schema

    @property
    def config(self):
        return self._config

    @property
    def config_path(self):
        if not self.in_memory:
            return os.path.join(self._repository, self.__config_filename)

    def validate(self):
        _config_validator(self._config, self._schema)

    def to_yaml(self):
        return yaml_config_dump(self._config,    # pragma: no cover
                                Dumper=SafeDumper,
                                default_flow_style=False)

    def load(self, storage_config=None):
        if self.in_memory:
            storage_config = MemoryConfig
        elif (storage_config is not None and
                not (inspect.isclass(storage_config) and
                     issubclass(storage_config, StorageConfig))):
            raise TypeError("Need a subclass of 'StorageConfig'")

        storage_config = storage_config or SQLiteConfig

        if self.exists():
            # Ignore param `storage_config`
            with open(self.config_path, "r") as stream:
                self._config = yaml_config_load(stream, SafeLoader)
                self._schema = self._get_storage_config_schema()
                self.validate()
        else:
            self._config = yaml_config_load(storage_config.config, SafeLoader)
            self._schema = storage_config.schema

        return self

    def save(self):
        if self.in_memory:
            return None

        if not os.path.isdir(self._repository):
            os.makedirs(self._repository)

        self.validate()
        with open(self.config_path, "w") as stream:
            yaml_config_dump(self._config,
                             stream,
                             Dumper=SafeDumper,
                             default_flow_style=False)

    def drop(self):
        if self.exists():
            os.remove(self.config_path)

    def exists(self):
        if not self.in_memory:
            return os.path.isfile(self.config_path)



from abc import ABCMeta, abstractmethod

from ..engine.helpers import with_metaclass


class StorageConfig(object):
    """Base class of storage config"""
    config = NotImplemented
    schema = NotImplemented


class AbstractStorage(with_metaclass(ABCMeta, object)):
    """
    """

    def __init__(self, repository, storage_config):
        self.is_opened = True
        self._repository = repository
        self._config = storage_config

    def __repr__(self):
        return "MontyStorage(engine={!r})".format(self.__class__.__name__)

    def __getattribute__(self, attr):
        def obj_attr(attr_):
            return object.__getattribute__(self, attr_)

        if not (attr == "is_opened" or obj_attr("is_opened")):
            # Run re-open if not checking open status nor is opened.
            obj_attr("_re_open")()

        return obj_attr(attr)

    def __getattr__(self, attr):
        return self._delegate(attr)

    def _delegate(self, attr):
        def to_db_level(db_name, *args, **kwargs):
            db = self.db_cls(db_name, self)
            return getattr(db, attr)(*args, **kwargs)
        return to_db_level

    def _re_open(self):
        """Auto re-open"""
        self.is_opened = True
        self._config.reload(repository=self._repository)

    def close(self):
        """Could do some clean up"""
        self.is_opened = False

    def wconcern_parser(self, client_kwargs):
        """
        Parsing storage specific write concern

        Optional, use Monty WriteConcern by default.
        Recive MontyClient kwargs, should parse kwargs and return a instance
        of `montydb.base.WriteConcern` class.
        """
        pass

    @property
    def repository(self):
        return self._repository

    @property
    def db_cls(self):
        raise NotImplementedError("")

    @abstractmethod
    def database_create(self):
        return NotImplemented

    @abstractmethod
    def database_drop(self):
        return NotImplemented

    @abstractmethod
    def database_list(self):
        return NotImplemented


class AbstractDatabase(with_metaclass(ABCMeta, object)):

    def __init__(self, name, storage):
        self._name = name
        self._storage = storage

    def __getattr__(self, attr):
        return self._delegate(attr)

    def _delegate(self, attr):
        def to_collection_level(collection_name,
                                write_concern,
                                codec_options,
                                *args,
                                **kwargs):
            collection = self.col_cls(collection_name,
                                      self,
                                      write_concern,
                                      codec_options)
            return getattr(collection, attr)(*args, **kwargs)
        return to_collection_level

    @property
    def col_cls(self):
        raise NotImplementedError("")

    @abstractmethod
    def collection_exists(self):
        return NotImplemented

    @abstractmethod
    def collection_create(self):
        return NotImplemented

    @abstractmethod
    def collection_drop(self):
        return NotImplemented

    @abstractmethod
    def collection_list(self):
        return NotImplemented


class AbstractCollection(with_metaclass(ABCMeta, object)):

    def __init__(self, name, database, write_concern, codec_options):
        self._name = name
        self._database = database
        self.wconcern = write_concern
        self.coptions = codec_options

    def __getattr__(self, attr):
        return self._delegate(attr)

    def _delegate(self, attr):
        def to_cursor_level(*args, **kwargs):
            cursor = self.cursor_cls(self)
            return getattr(cursor, attr)(*args, **kwargs)
        return to_cursor_level

    @property
    def cursor_cls(self):
        raise NotImplementedError("")

    @abstractmethod
    def write_one(self):
        return NotImplemented

    @abstractmethod
    def write_many(self):
        return NotImplemented


class AbstractCursor(with_metaclass(ABCMeta, object)):

    def __init__(self, collection):
        self._collection = collection

    @abstractmethod
    def query(self):
        return NotImplemented



# Copyright 2011-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Modifications copyright (C) 2018 davidlatwe
#
# Assembling crucial classes and functions form pymongo module,
# some of them may modified by needs.


import collections

from bson import BSON, SON
from bson.py3compat import integer_types, string_type, abc, iteritems
from bson.codec_options import CodecOptions, DEFAULT_CODEC_OPTIONS
from bson.raw_bson import RawBSONDocument
from bson.codec_options import _parse_codec_options


ASCENDING = 1
DESCENDING = -1


def _bson_touch(doc, codec_op=DEFAULT_CODEC_OPTIONS):
    """
    BSON will convert bytes string `b""` to unicode in PY2 since it's also
    `str` type, while in PY3 keep it as `bytes`. This might cause trouble
    dealing `Binary` type querying.
    Inorder to help this kind of scenario, make MontyDB to behave more like
    MongoDB, the command doucment like cursor filter should be encoded to
    BSON and decode back before parsing it, just like PyMongo encoded to
    BSON before sending to server.
    """
    if isinstance(doc, abc.Mapping):
        # Preserve the order of the query document when decode it back
        order_keep = CodecOptions(document_class=SON)
        return BSON(BSON.encode(doc, False, codec_op)).decode(order_keep)
    else:
        return doc  # should be None type.


def validate_is_document_type(option, value):
    """Validate the type of method arguments that expect a MongoDB document."""
    if not isinstance(value, (collections.MutableMapping, RawBSONDocument)):
        raise TypeError("%s must be an instance of dict, bson.son.SON, "
                        "bson.raw_bson.RawBSONDocument, or "
                        "a type that inherits from "
                        "collections.MutableMapping" % (option,))


def validate_boolean(option, value):
    """Validates that 'value' is True or False."""
    if isinstance(value, bool):
        return value
    raise TypeError("%s must be True or False" % (option,))


def validate_is_mapping(option, value):
    """Validate the type of method arguments that expect a document."""
    if not isinstance(value, abc.Mapping):
        raise TypeError("%s must be an instance of dict, bson.son.SON, or "
                        "other type that inherits from "
                        "collections.Mapping" % (option,))


def _fields_list_to_dict(fields, option_name):
    """Takes a sequence of field names and returns a matching dictionary.
    ["a", "b"] becomes {"a": 1, "b": 1}
    and
    ["a.b.c", "d", "a.c"] becomes {"a.b.c": 1, "d": 1, "a.c": 1}
    """
    if isinstance(fields, abc.Mapping):
        return fields

    if isinstance(fields, (abc.Sequence, abc.Set)):
        if not all(isinstance(field, string_type) for field in fields):
            raise TypeError("%s must be a list of key names, each an "
                            "instance of %s" % (option_name,
                                                string_type.__name__))
        return dict.fromkeys(fields, 1)

    raise TypeError("%s must be a mapping or "
                    "list of key names" % (option_name,))


def _index_list(key_or_list, direction=None):
    """Helper to generate a list of (key, direction) pairs.

    Takes such a list, or a single key, or a single key and direction.
    """
    if direction is not None:
        return [(key_or_list, direction)]
    else:
        if isinstance(key_or_list, string_type):
            return [(key_or_list, ASCENDING)]
        elif not isinstance(key_or_list, (list, tuple)):
            raise TypeError("if no direction is specified, "
                            "key_or_list must be an instance of list")
        return key_or_list


def _index_document(index_list):
    """Helper to generate an index specifying document.

    Takes a list of (key, direction) pairs.
    """
    if isinstance(index_list, abc.Mapping):
        raise TypeError("passing a dict to sort/create_index/hint is not "
                        "allowed - use a list of tuples instead. did you "
                        "mean %r?" % list(iteritems(index_list)))
    elif not isinstance(index_list, (list, tuple)):
        raise TypeError("must use a list of (key, direction) pairs, "
                        "not: " + repr(index_list))
    if not len(index_list):
        raise ValueError("key_or_list must not be the empty list")

    index = SON()
    for (key, value) in index_list:
        if not isinstance(key, string_type):
            raise TypeError("first item in each key pair must be a string")
        if not isinstance(value, (string_type, int, abc.Mapping)):
            raise TypeError("second item in each key pair must be 1, -1, "
                            "'2d', 'geoHaystack', or another valid MongoDB "
                            "index specifier.")
        index[key] = value
    return index


class WriteConcern(object):
    """MontyWriteConcern
    """

    __slots__ = ("_document")

    def __init__(self, wtimeout=None):
        self._document = {}

        if wtimeout is not None:
            if not isinstance(wtimeout, integer_types):
                raise TypeError("wtimeout must be an integer")
            self._document["wtimeout"] = wtimeout

    @property
    def document(self):
        """The document representation of this write concern.
        """
        return self._document.copy()

    def __repr__(self):
        return ("MontyWriteConcern({})".format(
            ", ".join("%s=%s" % kvt for kvt in self.document.items()),))

    def __eq__(self, other):
        return self.document == other.document

    def __ne__(self, other):
        return self.document != other.document

    def __bool__(self):
        return bool(self.document)


def _parse_write_concern(options):
    """Parse write concern options."""
    wtimeout = options.get('wtimeout')
    return WriteConcern(wtimeout)


class ClientOptions(object):
    """ClientOptions"""

    def __init__(self, options, storage_wconcern=None):
        self.__options = options
        self.__codec_options = _parse_codec_options(options)

        if storage_wconcern is not None:
            self.__write_concern = storage_wconcern
        else:
            self.__write_concern = _parse_write_concern(options)

    @property
    def _options(self):
        """The original options used to create this ClientOptions."""
        return self.__options  # pragma: no cover

    @property
    def codec_options(self):
        """A :class:`~bson.codec_options.CodecOptions` instance."""
        return self.__codec_options

    @property
    def write_concern(self):
        """A :class:`~montydb.base.WriteConcern` instance."""
        return self.__write_concern


class BaseObject(object):
    """A base class that provides attributes and methods common
    to multiple montydb classes.

    SHOULD NOT BE USED BY DEVELOPERS EXTERNAL TO MONTYDB.
    """

    def __init__(self, codec_options, write_concern):

        if not isinstance(codec_options, CodecOptions):
            raise TypeError("codec_options must be an "  # pragma: no cover
                            "instance of bson.codec_options.CodecOptions")
        self.__codec_options = codec_options

        if not isinstance(write_concern, WriteConcern):
            raise TypeError("write_concern must be an "  # pragma: no cover
                            "instance of montydb.base.WriteConcern")
        self.__write_concern = write_concern

    @property
    def codec_options(self):
        """Read only access to the :class:`~bson.codec_options.CodecOptions`
        of this instance.
        """
        return self.__codec_options

    @property
    def write_concern(self):
        """Read only access to the :class:`~montydb.base.WriteConcern`
        of this instance.
        """
        return self.__write_concern
